"""ORM signal handlers."""

from django import dispatch
from django.db.models import Model
from django.db.models import signals as model_signals

from resolwe.permissions.models import Permission, PermissionObject

from .models import Observer
from .protocol import ChangeType, post_permission_changed, pre_permission_changed

# Global 'in migrations' flag to ignore signals during migrations.
# Signal handlers that access the database can crash the migration process.
IN_MIGRATIONS = False


def observe_containers(instance: Model):
    """Send notifications to containers when instance changes."""
    for container_attribute in ("collection", "entity"):
        container = getattr(instance, container_attribute, None)
        if container is not None:
            Observer.observe_instance_changes(container, ChangeType.UPDATE)


@dispatch.receiver(model_signals.pre_migrate)
def model_pre_migrate(*args, **kwargs):
    """Set 'in migrations' flag."""
    global IN_MIGRATIONS
    IN_MIGRATIONS = True


@dispatch.receiver(model_signals.post_migrate)
def model_post_migrate(*args, **kwargs):
    """Clear 'in migrations' flag."""
    global IN_MIGRATIONS
    IN_MIGRATIONS = False


@dispatch.receiver(pre_permission_changed)
def prepare_permission_change(instance, **kwargs):
    """Store old permissions for an object whose permissions are about to change."""
    if not IN_MIGRATIONS:
        instance._old_viewers = instance.users_with_permission(
            Permission.VIEW, with_superusers=True
        )


@dispatch.receiver(post_permission_changed)
def handle_permission_change(instance, **kwargs):
    """Compare permissions for an object whose permissions changed."""
    if not IN_MIGRATIONS:
        new = set(instance.users_with_permission(Permission.VIEW, with_superusers=True))
        # The "_old_viewers" property may not exist. For instance if data object is
        # created in the collection and collection permissions are assigned to it
        # without set_permission call.
        old = set(getattr(instance, "_old_viewers", []))
        gains = new - old
        losses = old - new
        Observer.observe_permission_changes(instance, gains, losses)
        observe_containers(instance)


@dispatch.receiver(model_signals.post_save)
def observe_model_modification(
    sender: type, instance: Model, created: bool = False, **kwargs
):
    """Receive model updates."""

    # Create signals will be caught when the PermissionModel is added.
    if created:
        return

    if isinstance(instance, PermissionObject) and not IN_MIGRATIONS:
        Observer.observe_instance_changes(instance, ChangeType.UPDATE)
        observe_containers(instance)


@dispatch.receiver(model_signals.pre_delete)
def observe_model_deletion(sender: type, instance: Model, **kwargs):
    """Receive model deletions."""
    if isinstance(instance, PermissionObject) and not IN_MIGRATIONS:
        Observer.observe_instance_changes(instance, ChangeType.DELETE)



"""Collection viewset."""
from django.db import transaction
from django.db.models import Prefetch

from rest_framework import exceptions, mixins, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.flow.filters import CollectionFilter
from resolwe.flow.models import Collection, DescriptorSchema
from resolwe.flow.models.annotations import AnnotationValue
from resolwe.flow.serializers import CollectionSerializer
from resolwe.flow.serializers.annotations import AnnotationFieldDictSerializer
from resolwe.observers.mixins import ObservableMixin
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.models import Permission, PermissionModel

from .mixins import (
    ParametersMixin,
    ResolweCheckSlugMixin,
    ResolweCreateModelMixin,
    ResolweUpdateModelMixin,
)


class BaseCollectionViewSet(
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    ResolwePermissionsMixin,
    ResolweCheckSlugMixin,
    ParametersMixin,
    viewsets.GenericViewSet,
):
    """Base API view for :class:`Collection` objects."""

    qs_descriptor_schema = DescriptorSchema.objects.select_related("contributor")
    qs_permission_model = PermissionModel.objects.select_related("user", "group")

    queryset = Collection.objects.select_related("contributor").prefetch_related(
        "data", Prefetch("descriptor_schema", queryset=qs_descriptor_schema)
    )

    filter_class = CollectionFilter
    permission_classes = (get_permissions_class(),)

    ordering_fields = (
        "contributor",
        "contributor__first_name",
        "contributor__last_name",
        "created",
        "id",
        "modified",
        "name",
    )
    ordering = "id"

    def get_queryset(self):
        """Prefetch permissions for current user."""
        return self.prefetch_current_user_permissions(self.queryset)

    def create(self, request, *args, **kwargs):
        """Only authenticated users can create new collections."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        return super().create(request, *args, **kwargs)

    @action(detail=False, methods=["post"])
    def duplicate(self, request, *args, **kwargs):
        """Duplicate (make copy of) ``Collection`` models."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        ids = self.get_ids(request.data)
        queryset = Collection.objects.filter(id__in=ids).filter_for_user(
            request.user, Permission.VIEW
        )
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Collections with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        duplicated = queryset.duplicate(contributor=request.user)

        serializer = self.get_serializer(duplicated, many=True)
        return Response(serializer.data)


class CollectionViewSet(ObservableMixin, BaseCollectionViewSet):
    """API view for :class:`Collection` objects."""

    serializer_class = CollectionSerializer

    def get_queryset(self):
        """Prefetch entity data and return queryset."""
        return super().get_queryset().prefetch_related("entity_set")

    @action(detail=True, methods=["post"])
    def add_fields_to_collection(self, request, pk=None):
        """Add the given list of AnnotaitonFields to the given collection."""
        # No need to check for permissions, since post requires edit by default.
        collection = self.get_object()
        # Read and validate the request data.
        serializer = AnnotationFieldDictSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        # Add the fields to the collection.
        collection.annotation_fields.add(
            *serializer.validated_data["annotation_fields"]
        )
        return Response()

    @action(detail=True, methods=["post"])
    def remove_fields_from_collection(self, request, pk=None):
        """Remove the given list of AnnotaitonFields from the given collection.

        The required argument is annotation_fields and optional confirm_action.
        """
        # No need to check for permissions, since post requires edit by default.
        collection = self.get_object()
        serializer = AnnotationFieldDictSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        if not serializer.validated_data["confirm_action"]:
            raise exceptions.ValidationError(
                "Annotations for the given fields will be removed from the samples in "
                "the collection. Set 'confirm_action' argument to 'True' to confirm."
            )

        annotation_fields = serializer.validated_data["annotation_fields"]
        # Delete annotation values from the samples in this collection and remove the
        # annotation fields from the collection.
        with transaction.atomic():
            AnnotationValue.objects.filter(
                entity__collection=collection, field__in=annotation_fields
            ).delete()
            collection.annotation_fields.remove(*annotation_fields)
        return Response()
